# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implementation of Cluster Resolvers for GCE Instance Groups."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from tensorflow.contrib.cluster_resolver.python.training.cluster_resolver import ClusterResolver
from tensorflow.python.training.server_lib import ClusterSpec

_GOOGLE_API_CLIENT_INSTALLED = True
try:
  from googleapiclient import discovery  # pylint: disable=g-import-not-at-top
  from oauth2client.client import GoogleCredentials  # pylint: disable=g-import-not-at-top
except ImportError:
  _GOOGLE_API_CLIENT_INSTALLED = False


class GceClusterResolver(ClusterResolver):
  """Cluster Resolver for Google Compute Engine.

  This is an implementation of cluster resolvers for the Google Compute Engine
  instance group platform. By specifying a project, zone, and instance group,
  this will retrieve the IP address of all the instances within the instance
  group and return a Cluster Resolver object suitable for use for distributed
  TensorFlow.
  """

  def __init__(self,
               project,
               zone,
               instance_group,
               port,
               job_name='worker',
               credentials='default',
               service=None):
    """Creates a new GceClusterResolver object.

    This takes in a few parameters and creates a GceClusterResolver project. It
    will then use these parameters to query the GCE API for the IP addresses of
    each instance in the instance group.

    Args:
      project: Name of the GCE project
      zone: Zone of the GCE instance group
      instance_group: Name of the GCE instance group
      port: Port of the listening TensorFlow server (default: 8470)
      job_name: Name of the TensorFlow job this set of instances belongs to
      credentials: GCE Credentials. If nothing is specified, this defaults to
        GoogleCredentials.get_application_default()
      service: The GCE API object returned by the googleapiclient.discovery
        function. (Default: discovery.build('compute', 'v1')). If you specify a
        custom service object, then the credentials parameter will be ignored.

    Raises:
      ImportError: If the googleapiclient is not installed.
    """
    self._project = project
    self._zone = zone
    self._instance_group = instance_group
    self._job_name = job_name
    self._port = port
    self._credentials = credentials

    if credentials == 'default':
      if _GOOGLE_API_CLIENT_INSTALLED:
        self._credentials = GoogleCredentials.get_application_default()

    if service is None:
      if not _GOOGLE_API_CLIENT_INSTALLED:
        raise ImportError('googleapiclient must be installed before using the '
                          'GCE cluster resolver')
      self._service = discovery.build(
          'compute', 'v1',
          credentials=self._credentials)
    else:
      self._service = service

  def cluster_spec(self):
    """Returns a ClusterSpec object based on the latest instance group info.

    This returns a ClusterSpec object for use based on information from the
    specified instance group. We will retrieve the information from the GCE APIs
    every time this method is called.

    Returns:
      A ClusterSpec containing host information retrieved from GCE.
    """
    request_body = {'instanceState': 'RUNNING'}
    request = self._service.instanceGroups().listInstances(
        project=self._project,
        zone=self._zone,
        instanceGroups=self._instance_group,
        body=request_body,
        orderBy='name')

    worker_list = []

    while request is not None:
      response = request.execute()

      items = response['items']
      for instance in items:
        instance_name = instance['instance'].split('/')[-1]

        instance_request = self._service.instances().get(
            project=self._project,
            zone=self._zone,
            instance=instance_name)

        if instance_request is not None:
          instance_details = instance_request.execute()
          ip_address = instance_details['networkInterfaces'][0]['networkIP']
          instance_url = '%s:%s' % (ip_address, self._port)
          worker_list.append(instance_url)

      request = self._service.instanceGroups().listInstances_next(
          previous_request=request,
          previous_response=response)

    worker_list.sort()
    return ClusterSpec({self._job_name: worker_list})

  def master(self):
    return ''



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Special functions that only make sense for AutoGraph.

These functions are meant to ensure feature parity between Python and AutoGraph,
so that the exact same code works in both modes. In general, AutoGraph will
replace these calls.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.contrib.autograph.operators import data_structures


def stack(list_or_tensor, element_dtype=None, strict=True):
  """Stacks the input, if it admits the notion of stacking.

  For example, a list of tensors can be stacked into a larger tensor. This
  function is similar to tf.stack, but it accepts non-lists and lists of
  non-tensors as arguments. In the latter case, the function does nothing.

  Args:
    list_or_tensor: Any
    element_dtype: tf.DType, optional dtypedtype for the elements in the list.
        Required if the input is stackable, and the list is untyped.
    strict: bool, if True an error is raised if the input is not stackable.
        Otherwise the function is a no-op.

  Returns:
    Any, if the input is stackable, the result will be a tf.Tensor. Otherwise,
    if strict=False, the result will be list_or_tensor.

  Raises:
    ValueError: if strict=True and the input is not stackable.
  """
  if strict:
    def raise_error(x):
      raise ValueError('%s must be stackable when strict=True' % x)
    original_call = raise_error
  else:
    original_call = lambda x: x
  return data_structures.list_stack(
      list_or_tensor,
      data_structures.ListStackOpts(
          element_dtype=element_dtype, original_call=original_call))



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Container for origin source code information before AutoGraph compilation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import gast

from tensorflow.contrib.autograph.pyct import anno
from tensorflow.python.util import tf_inspect


class CodeLocation(
    collections.namedtuple('CodeLocation', ('file_path', 'line_number'))):
  """Location of a line of code.

  Attributes:
    file_path: text, the full path to the file containing the code.
    line_number: Int, the 1-based line number of the code in its file.
  """
  pass


class OriginInfo(
    collections.namedtuple('OriginInfo',
                           ('file_path', 'function_name', 'line_number',
                            'column_offset', 'source_code_line'))):
  """Container for information about the source code before conversion.

  Instances of this class contain information about the source code that
  transformed code originated from. Examples include:
    * line number
    * file name
    * original user code
  """

  def as_frame(self):
    """Makes a traceback frame tuple.

    Returns:
      A tuple of (file_path, line_number, function_name, source_code_line).
    """
    return (self.file_path, self.line_number, self.function_name,
            self.source_code_line)


# TODO(znado): Consider refactoring this into a Visitor.
def resolve(node, source, function=None):
  """Adds an origin information to all nodes inside the body of function.

  Args:
    node: The AST node for the function whose body nodes will be annotated.
    source: Text, the source code string for the function whose body nodes will
      be annotated.
    function: Callable, the function that will have all nodes inside of it
      annotation with an OriginInfo annotation with key anno.Basic.ORIGIN.  If
      it is None then only the line numbers and column offset will be set in the
      annotation, with the rest of the information being None.

  Returns:
    A tuple of the AST node for function and a String containing its source
    code.
  """
  if function:
    _, function_lineno = tf_inspect.getsourcelines(function)
    function_filepath = tf_inspect.getsourcefile(function)
  else:
    function_lineno = None
    function_filepath = None
  source_lines = source.split('\n')
  for n in gast.walk(node):
    if hasattr(n, 'lineno'):
      # n.lineno is relative to the start of the enclosing function, so need to
      # offset it by the line of the function.
      source_code_line = source_lines[n.lineno - 1]
      if function:
        source_lineno = n.lineno + function_lineno - 1
        function_name = function.__name__
      else:
        source_lineno = n.lineno
        function_name = None
      anno.setanno(
          n, anno.Basic.ORIGIN,
          OriginInfo(function_filepath, function_name, source_lineno,
                     n.col_offset, source_code_line))



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Error rewriting logic.

Contains the functions responsible for rewriting tracebacks of errors raised
in AutoGraph (AG) code to refer to user written code, so that errors only refer
to the original user code.

When 'user code' is used in comments it refers to the original source code that
the user wrote and is converting using AutoGraph.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import logging
import sys
import traceback

from tensorflow.contrib.autograph.pyct.origin_info import CodeLocation
from tensorflow.python.framework import errors_impl
from tensorflow.python.util import tf_inspect


class GraphConstructionError(Exception):
  """Error for graph construction errors from AutoGraph generated code."""

  def __init__(self, original_error, custom_traceback):
    self.original_error = original_error
    self.custom_traceback = custom_traceback
    super(GraphConstructionError, self).__init__()

  def __str__(self):
    traceback_str = ''.join(traceback.format_list(self.custom_traceback))
    return ('Traceback (most recent call last):\n' + traceback_str + '\n' + str(
        self.original_error) + '\n')


class TfRuntimeError(Exception):
  """Error wrapper for runtime errors raised by AutoGraph generated code."""

  def __init__(self, op_name, op_message, custom_traceback):
    self.op_name = op_name
    self.op_message = op_message
    self.custom_traceback = custom_traceback
    super(TfRuntimeError, self).__init__()

  def __str__(self):
    message = '%s\n\nCaused by op %r, defined at:\n' % (self.op_message,
                                                        self.op_name)
    return message + ''.join(traceback.format_list(self.custom_traceback))


def _rewrite_frame(source_map, cleaned_traceback, stack_frame_indices):
  """Rewrites the stack frames at the given indices using the given source map.

  Args:
    source_map: Dict[CodeLocation, OriginInfo], a mapping between the user and
        AG generated code.
    cleaned_traceback: List[Tuple[text, text, text, text]], the current
        traceback.
    stack_frame_indices: Iterable[Int], frame indices to possibly rewrite if
        there are matching source mapping keys.

  Returns:
    None
  """
  for frame_index in stack_frame_indices:
    # (file_path, line number, function name, code)
    file_path, line_number, _, _ = cleaned_traceback[frame_index]
    source_map_key = CodeLocation(file_path=file_path, line_number=line_number)
    found_mapping = source_map_key in source_map
    if found_mapping:
      cleaned_traceback[frame_index] = source_map[source_map_key].as_frame()


# TODO(znado): Make more robust to name changes in the rewriting logic.
def _remove_rewrite_frames(tb):
  """Remove stack frames containing the error rewriting logic."""
  cleaned_tb = []
  for f in tb:
    if 'ag__.rewrite_graph_construction_error' not in f[3]:
      cleaned_tb.append(f)
  return cleaned_tb


def rewrite_graph_construction_error(source_map):
  """Rewrites errors raised by non-AG APIs inside AG generated code.

  Meant to be called from the try/except block inside each AutoGraph generated
  function.  Only rewrites the traceback frames corresponding to the function
  that this is called from.  When we raise a GraphConstructionError at the end
  it is then caught by calling functions, where they can be responsible for
  rewriting their own frames.

  Args:
    source_map: Dict[CodeLocation, OriginInfo], a mapping between the user and
        AG generated code.

  Raises:
    GraphConstructionError: The rewritten underlying error.
    Exception: The underlying error, if it could not be rewritten.
  """
  error_info = sys.exc_info()
  _, original_error, e_traceback = error_info
  assert original_error is not None
  try:
    _, _, _, func_name, _, _ = tf_inspect.stack()[1]
    # The latest function call is added to the beginning of a traceback, but
    # when rewriting the traceback of multiple function calls the previous
    # functions' except blocks may have already rewritten their own frames so
    # we want to copy over all of the previous frames. We may have rewritten
    # previous frames only if the error is a GraphConstructionError.
    if isinstance(original_error, GraphConstructionError):
      cleaned_traceback = traceback.extract_tb(e_traceback)
      previous_traceback = original_error.custom_traceback
      cleaned_traceback = [cleaned_traceback[0]] + previous_traceback
    else:
      cleaned_traceback = traceback.extract_tb(e_traceback)
    cleaned_traceback = _remove_rewrite_frames(cleaned_traceback)

    current_frame_indices = []
    # This code is meant to be called from the try/except block that wraps a
    # function body.  Here we look for all frames that came from the function
    # that this wraps, look for any matching line numbers in the source
    # mapping, and then rewrite them if matches are found.
    for fi, frame in enumerate(cleaned_traceback):
      _, _, frame_func_name, _ = frame
      if frame_func_name == func_name:
        current_frame_indices.append(fi)
        break
    if current_frame_indices:
      _rewrite_frame(source_map, cleaned_traceback, current_frame_indices)

    if isinstance(original_error, GraphConstructionError):
      original_error.custom_traceback = cleaned_traceback
      new_error = original_error
    else:
      new_error = GraphConstructionError(original_error, cleaned_traceback)
  except Exception:
    logging.exception('Error while rewriting AutoGraph error:')
    raise original_error
  else:
    raise new_error
  finally:
    # Addresses warning https://docs.python.org/2/library/sys.html#sys.exc_info.
    del e_traceback


def rewrite_tf_runtime_error(error, source_map):
  """Rewrites TensorFlow runtime errors raised by ops created in AG code.

  Args:
    error: error_impl.OpError, an TensorFlow error that will have its traceback
        rewritten.
    source_map: Dict[CodeLocation, OriginInfo], a mapping between the user and
        AG generated code.

  Returns:
    A TfRuntimeError with a traceback rewritten according to the given
    source mapping.
  """
  # Check for cases where we leave a user method and re-enter it in the
  # traceback.  This is done by looking at the function names when the
  # filenames are from any files the user code is in.  If we find a case where
  # we return to a user method after leaving it then we cut out the frames in
  # between because we assume this means these in between frames are from
  # internal AutoGraph code that shouldn't be included.
  #
  # An example of this is:
  #
  #  File "file1.py", line 57, in my_func
  #    ...
  #  File "control_flow_ops.py", line 231, in cond
  #    ...
  #  File "control_flow_ops.py", line 1039, in inner_cond
  #    ...
  #  File "file1.py", line 68, in my_func
  #    ...
  #
  # Where we would remove the control_flow_ops.py frames because we re-enter
  # my_func in file1.py.
  #
  # The source map keys are (file_path, line_number) so get the set of all user
  # file_paths.
  try:
    all_user_files = set(k.file_path for k in source_map)
    cleaned_traceback = []
    last_user_frame_index = None
    last_user_user_file_path = None
    last_user_user_fn_name = None
    for fi, frame in enumerate(error.op.traceback):
      frame_file_path, frame_line_number, _, _ = frame
      src_map_key = CodeLocation(
          file_path=frame_file_path, line_number=frame_line_number)
      if frame_file_path in all_user_files:
        if src_map_key in source_map:
          original_fn_name = source_map[src_map_key].function_name
          if (last_user_frame_index is not None and
              last_user_user_file_path == frame_file_path):
            if last_user_user_fn_name == original_fn_name:
              cleaned_traceback = cleaned_traceback[:last_user_frame_index]
            else:
              cleaned_traceback = cleaned_traceback[:last_user_frame_index + 1]
          last_user_user_fn_name = original_fn_name
        else:
          last_user_user_fn_name = None
        last_user_frame_index = fi
        last_user_user_file_path = frame_file_path
      cleaned_traceback.append(frame)

    for fi in range(len(cleaned_traceback)):
      _rewrite_frame(source_map, cleaned_traceback, [fi])
    op_name = error.op.name
    op_message = error.message
    rewritten_error = TfRuntimeError(op_name, op_message, cleaned_traceback)
    return rewritten_error
  except Exception:  # pylint: disable=broad-except
    logging.exception('Error while rewriting AutoGraph error:')
    return error


# TODO(znado): Add arg to enable different levels of error rewriting.
@contextlib.contextmanager
def improved_errors(converted_function):
  """Context manager that rewrites runtime errors.

  This context manager will rewrite runtime errors so that their traceback
  is relative to the original code before conversion.

  Use with the output of to_graph, and wrap the execution of respective ops.
  Example:

    converted_my_func = ag.to_graph(my_func)
    ops = converted_my_func(...)

    with ag.improved_errors(converted_my_func):
      sess.run(ops)

  Args:
    converted_function: Callable[..., Any], the output of a to_graph call

  Yields:
    None

  Raises:
    TfRuntimeError: if any OpError originates in the converted code, it will
        be wrapped into a TfRuntimeError
    ValueError: If converted_function is not generated by AutoGraph
  """
  if (getattr(converted_function, 'ag_source_map', None) is None or
      not converted_function.ag_source_map):
    raise ValueError(
        'converted_function must be the result of an autograph.to_graph call')
  try:
    yield
  except errors_impl.OpError as e:
    raise rewrite_tf_runtime_error(e, converted_function.ag_source_map)
